import unittest

from pathlib import Path

from qgis_plugin_manager.local_directory import LocalDirectory


class PluginManager(unittest.TestCase):

    def setUp(self) -> None:
        self.local = LocalDirectory(Path('fixtures/plugins'))

    def test_list_existing_plugins(self):
        """ Test to find existing plugins. """
        self.assertListEqual(['plugin_a', 'plugin_b'], self.local.plugins())
        self.assertListEqual(['missing_init', 'missing_metadata'], self.local.invalid)

    def test_read_metadata(self):
        self.assertIsNone(self.local.plugin_metadata('do_not_exist', 'author'))
        self.assertIsNone(self.local.plugin_metadata('missing_init', 'author'))
        self.assertIsNone(self.local.plugin_metadata('missing_metadata', 'author'))

        self.assertEqual('Hazel Nutt', self.local.plugin_metadata('plugin_a', 'author'))
        self.assertEqual('', self.local.plugin_metadata('plugin_a', 'do_not_exist'))

        self.assertListEqual(
            ['Plugin A', '1.0.0', '3.0', '', 'Hazel Nutt'],
            self.local.plugin_all_info('plugin_a'))


if __name__ == '__main__':
    unittest.main()



import configparser

from pathlib import Path
from typing import List, Union

from .remote import Remote
from .utils import pretty_table, parse_version


class LocalDirectory:

    def __init__(self, folder: Path, qgis_version: str = None):
        self.folder = folder
        self._plugins = None
        self._invalid = []

        self.qgis_version = None
        if qgis_version:
            self.qgis_version = qgis_version.split('.')
            if len(self.qgis_version) != 3:
                self.qgis_version = None
            else:
                self.qgis_version = [int(i) for i in self.qgis_version]

    def init(self):
        source_file = self.folder.joinpath('sources.list')
        if source_file.exists():
            print("sources.list already existing. Quit")
            exit(1)

        if self.qgis_version:
            version = f"{self.qgis_version[0]}.{self.qgis_version[1]}"
            print(f"Init https://plugins.qgis.org with {version}")
            server = f"https://plugins.qgis.org/plugins/plugins.xml?qgis={version}\n"
        else:
            print(f"QGIS version is unknown, creating with a default 3.16")
            server = f"https://plugins.qgis.org/plugins/plugins.xml?qgis=3.16\n"

        with open(source_file, 'w', encoding='utf8') as f:
            f.write(server)

    def plugins(self) -> List[str]:
        self._plugins = []
        for folder in self.folder.iterdir():

            if not folder.is_dir():
                continue

            if folder.name.startswith('.'):
                continue

            have_python = list(folder.glob('*.py'))
            have_metadata = list(folder.glob('metadata.txt'))
            if have_python and have_metadata:
                self._plugins.append(folder.name)
            else:
                self._invalid.append(folder.name)

        # Weird issue, there are duplicates
        self._invalid = list(dict.fromkeys(self._invalid))

        return self._plugins

    def plugin_metadata(self, plugin: str, key: str) -> Union[str, None]:
        if self._plugins is None:
            self.plugins()

        if plugin not in self._plugins:
            return None

        config_parser = configparser.ConfigParser()

        with Path(self.folder / Path(f"{plugin}/metadata.txt")).open() as f:
            config_parser.read_file(f)

        try:
            return config_parser.get('general', key)
        except configparser.NoOptionError:
            return ''

    @property
    def invalid(self) -> list:
        return self._invalid

    def plugin_all_info(self, plugin: str) -> Union[List[str], None]:
        if self._plugins is None:
            self.plugins()

        if plugin not in self._plugins:
            return None

        data = []
        for item in ("name", "version", "qgisMinimumVersion", "qgisMaximumVersion", "author"):
            data.append(self.plugin_metadata(plugin, item))
        return data

    def print_table(self):
        if self._plugins is None:
            self.plugins()

        remote = Remote(self.folder)

        print(f"List all plugins in {self.folder.absolute()}\n")
        headers = ['Name', 'Version', 'QGIS min', 'QGIS max', 'Author', 'Action ⚠']
        headers = [f"  {i}  " for i in headers]
        data = []
        for plugin in self.plugins():
            plugin_data = self.plugin_all_info(plugin)

            latest = remote.latest(plugin)
            current = plugin_data[1]

            qgis_min = parse_version(plugin_data[2])
            qgis_max = parse_version(plugin_data[3])
            extra_info = []

            if latest:
                if latest.startswith('v'):
                    latest = latest[1:]

                if latest > current:
                    extra_info.append(f"Upgrade to {latest}")

                if self.qgis_version and qgis_min:
                    if qgis_min > self.qgis_version:
                        extra_info.append(f"QGIS Minimum {plugin_data[2]}")

                if self.qgis_version and qgis_max:
                    if qgis_max < self.qgis_version:
                        extra_info.append(f"QGIS Maximum {plugin_data[3]}")

            else:
                extra_info.append('Unkown version')

            plugin_data.append(';'.join(extra_info))
            data.append(plugin_data)
        print(pretty_table(data, headers))

        if len(self._invalid) >= 1:
            print(pretty_table(self._invalid, ['Invalid']))



#!/usr/bin/env python3

import argparse

from pathlib import Path

from qgis_plugin_manager.local_directory import LocalDirectory
from qgis_plugin_manager.remote import Remote
from qgis_plugin_manager.utils import qgis_server_version


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "-v", "--version", help="Print the version and exit", action="store_true"
    )

    subparsers = parser.add_subparsers(
        title="commands", description="qgis-plugin-manager command", dest="command"
    )

    subparsers.add_parser("init", help="Create the `sources.list` with plugins.qgis.org as remote")

    subparsers.add_parser("list", help="List all plugins in the directory")

    subparsers.add_parser("remote", help="List all remote server")

    subparsers.add_parser('update', help="Update all index files")

    cache = subparsers.add_parser("cache", help="Look for a plugin in the cache")
    cache.add_argument("plugin_name", help="The plugin to look for")

    install = subparsers.add_parser('install', help="Install a plugin")
    install.add_argument("plugin_name", help="The plugin to install, suffix '==version' is optional")

    args = parser.parse_args()

    # print the version and exit
    if args.version:
        import pkg_resources

        print(
            "qgis-plugin-manager version: {}".format(
                pkg_resources.get_distribution("qgis-plugin-manager").version
            )
        )
        qgis = qgis_server_version()
        if qgis:
            print(f"QGIS server version {qgis_server_version()}")
        parser.exit()

    # if no command is passed, print the help and exit
    if not args.command:
        parser.print_help()
        parser.exit()

    exit_val = 0

    if args.command == "update":
        remote = Remote(Path('.'))
        remote.update()
    elif args.command in ["list", "init"]:
        qgis = qgis_server_version()
        if qgis:
            print(f"QGIS server version {qgis}")
        else:
            print(f"QGIS server version unknown")

        plugins = LocalDirectory(Path('.'), qgis_version=qgis)

        if args.command == "list":
            plugins.print_table()
        else:
            plugins.init()

    elif args.command == "remote":
        remote = Remote(Path('.'))
        remote.print_list()

    elif args.command == "cache":
        remote = Remote(Path('.'))
        latest = remote.latest(args.plugin_name)
        if latest is None:
            print("Plugin not found")
        else:
            print(f"Plugin {args.plugin_name} : {latest} available")

    elif args.command == "install":
        remote = Remote(Path('.'))
        parameter = args.plugin_name.split('==')
        remote.install(*parameter)

    return exit_val


if __name__ == "__main__":
    exit(main())



import unittest

from pathlib import Path

from qgis_plugin_manager.remote import Remote


class PluginManager(unittest.TestCase):

    def setUp(self) -> None:
        self.remote = Remote(Path('fixtures/plugins'))

    def test_list_remote(self):
        self.assertListEqual(
            ["https://my.url/plugins.xml", "https://my.repo/plugins.xml"],
            self.remote.remote_list())

    @unittest.expectedFailure
    def test_latest_pgmetadata(self):
        # There is 0.1.0 and 0.2.0 in these XML files
        self.assertEqual('0.2.0', self.remote.latest("PgMetadata"))



import unittest

from qgis_plugin_manager.utils import parse_ldd, parse_version


class TestUtils(unittest.TestCase):

    def test_qgis_server_version(self):
        """ Trying the LDD command to know QGIS Server version. """
        output = (
            "	libqgis_server.so.3.10.4 => /lib/libqgis_server.so.3.10.4 (0x00007f6c76269000)\n"
            "	libqgis_core.so.3.10.4 => /lib/libqgis_core.so.3.10.4 (0x00007f6c74fd4000)"
        )
        self.assertEqual('3.10.4', parse_ldd(output))

        output = (
            "	libqgis_server.so.3.4.4 => /lib/libqgis_server.so.3.4.4 (0x00007f6c76269000)\n"
            "	libqgis_core.so.3.4.4 => /lib/libqgis_core.so.3.4.4 (0x00007f6c74fd4000)"
        )
        self.assertEqual('3.4.4', parse_ldd(output))

        output = (
            "libqgis_server.so.3.16.6 => /home/etienne/dev/app/qgis-stable/lib/libqgis_server.so.3.16.6 (0x00007f6b2dd30000)\n"
            "libqgis_core.so.3.16.6 => /home/etienne/dev/app/qgis-stable/lib/libqgis_core.so.3.16.6 (0x00007f6b2b879000)"
        )
        self.assertEqual('3.16.6', parse_ldd(output))

    def test_version(self):
        self.assertListEqual([3, 10, 0], parse_version("3.10"))
        self.assertListEqual([3, 10, 0], parse_version("3.10.0"))
        self.assertIsNone(parse_version(""))



import re
import shutil
import urllib
import urllib.request
import xml.etree.ElementTree as ET

from pathlib import Path

from qgis_plugin_manager.definitions import Plugin


class Remote:

    def __init__(self, folder: Path):
        self.folder = folder
        self.list = None
        self.list_plugins = None

    def remote_list(self) -> list:
        self.list = []
        source_list = Path(self.folder / 'sources.list')
        if not source_list.exists():
            return []

        with source_list.open() as f:
            for line in f.readlines():
                if not line.startswith('#'):
                    self.list.append(line.strip())
        return self.list

    def print_list(self):
        if self.list is None:
            self.remote_list()

        print("List of remotes :\n")
        print('\n'.join(self.list))

    def update(self):
        if self.list is None:
            self.remote_list()

        cache = Path(self.folder / ".cache_qgis_plugin_manager")
        if cache.exists():
            shutil.rmtree(cache)
        cache.mkdir()

        for i, server in enumerate(self.list):
            print(f"Downloading {server}...")
            request = urllib.request.Request(server, headers={'User-Agent': 'Mozilla/5.0'})
            try:
                f = urllib.request.urlopen(request)
            except urllib.error.HTTPError as e:
                print(f"\t{e}")
                continue

            filename = ""
            for x in server:
                if x.isalnum():
                    filename += x
                else:
                    filename += '-'

            filename = re.sub(r"\-+", "-", filename)
            with open(Path(cache / f"{filename}.xml"), 'wb') as output:
                output.write(f.read())

            print("\tOk")

    def latest(self, plugin_name):
        plugins = {}
        self.list_plugins = {}

        cache = Path(self.folder / ".cache_qgis_plugin_manager")
        for file in cache.iterdir():
            if not file.name.endswith('.xml'):
                continue

            tree = ET.parse(file.absolute())
            root = tree.getroot()
            for plugin in root:

                xml_plugin_name = plugin.attrib['name']
                if xml_plugin_name in plugins.keys():
                    previous_parsed_version = plugins[xml_plugin_name].split('.')
                    new_parsed_version = plugin.attrib['version'].split('.')
                    if previous_parsed_version < new_parsed_version:
                        plugins[xml_plugin_name] = plugin.attrib['version']
                    else:
                        continue
                else:
                    plugins[xml_plugin_name] = plugin.attrib['version']

                plugin_obj = Plugin()
                data = {}
                for element in plugin:
                    if element.tag in plugin_obj._fields:
                        data[element.tag] = element.text

                plugin_obj = Plugin(**data)
                self.list_plugins[xml_plugin_name] = plugin_obj

        return plugins.get(plugin_name)

    def install(self, plugin_name, version="latest"):
        print(f"Installation {plugin_name} {version}")

        xml_version = self.latest(plugin_name)
        if xml_version is None:
            return

        url = self.list_plugins[plugin_name].download_url
        file_name = self.list_plugins[plugin_name].file_name

        if version != 'latest':
            actual = self.list_plugins[plugin_name].version
            url = url.replace(actual, version)
            file_name = file_name.replace(actual, version)

        request = urllib.request.Request(url, headers={'User-Agent': 'Mozilla/5.0'})
        try:
            f = urllib.request.urlopen(request)
        except urllib.error.HTTPError:
            print(f"Plugin {plugin_name} {version} not found.")
            return

        zip_file = Path(self.folder / file_name)
        with open(zip_file, 'wb') as output:
            output.write(f.read())

        existing = Path(self.folder / plugin_name)
        if existing.exists():
            shutil.rmtree(existing)

        import zipfile
        with zipfile.ZipFile(zip_file, 'r') as zip_ref:
            zip_ref.extractall(self.folder)

        zip_file.unlink()

        print(f"\tOk {zip_file.name}")



import os
import shutil
import unittest

from pathlib import Path

from qgis_plugin_manager.local_directory import LocalDirectory
from qgis_plugin_manager.remote import Remote


@unittest.skipIf(os.getenv('CI') != 'true', "Only run on CI")
class FullInstallNetwork(unittest.TestCase):

    def setUp(self) -> None:
        self.plugin_name = 'QuickOSM'
        self.repository = "https://plugins.qgis.org/plugins/plugins.xml?qgis=3.10"
        self.directory = Path('fixtures/plugins')
        self.remote = Remote(self.directory)
        self.local = LocalDirectory(self.directory)

        self.plugin_path = Path(self.directory / self.plugin_name)
        if self.plugin_path.exists():
            shutil.rmtree(self.plugin_path)

        shutil.copy(Path(self.directory / 'sources.list'), Path(self.directory / 'sources.list.back'))

    def tearDown(self) -> None:
        if self.plugin_path.exists():
            shutil.rmtree(self.plugin_path)

        shutil.copy(Path(self.directory / 'sources.list.back'), Path(self.directory / 'sources.list'))

    def test_install_network(self):

        self.assertFalse(Path(self.directory / 'QuickOSM').exists())
        self.assertNotIn(self.plugin_name, self.local.plugins())

        with open(Path(self.directory / 'sources.list'), 'w') as file:
            file.write(self.repository)

        self.remote.update()

        version = '1.1.1'
        self.remote.install(self.plugin_name, version)
        self.assertTrue(Path(self.directory / 'QuickOSM').exists())
        self.local.plugins()
        self.assertEqual(version, self.local.plugin_metadata(self.plugin_name, 'version'))

        self.remote.install(self.plugin_name)
        self.assertTrue(Path(self.directory / 'QuickOSM').exists())
        self.assertNotEqual(version, self.local.plugin_metadata(self.plugin_name, 'version'))



import os
import re
import subprocess
from typing import Union


def pretty_table(iterable, header):
    """ Copy/paste from http://stackoverflow.com/a/40426743/2395485 """
    max_len = [len(x) for x in header]
    for row in iterable:
        row = [row] if type(row) not in (list, tuple) else row
        for index, col in enumerate(row):
            if max_len[index] < len(str(col)):
                max_len[index] = len(str(col))
    output = '-' * (sum(max_len) + 1) + '\n'
    output += '|' + ''.join(
        [h + ' ' * (l - len(h)) + '|' for h, l in zip(header, max_len)]) + '\n'
    output += '-' * (sum(max_len) + 1) + '\n'
    for row in iterable:
        row = [row] if type(row) not in (list, tuple) else row
        output += '|' + ''.join(
            [
                str(c) + ' ' * (
                    l - len(str(c))) + '|' for c, l in zip(
                    row, max_len)]) + '\n'
    output += '-' * (sum(max_len) + 1) + '\n'
    return output


def parse_version(version: str) -> Union[None, list]:
    if version is None or version == "":
        return None

    version = [int(i) for i in version.split(".")]
    if len(version) == 2:
        version.append(0)
    return version


def parse_ldd(command_output: str):
    if isinstance(command_output, str):
        command_output = command_output.split('\n')

    for line in command_output:
        # line = [i.strip() for i in line.split("=>")]

        if isinstance(line, list):
            line = line[0]

        exp = re.search(r"\d+.\d+.\d+", line)
        if exp:
            return exp.group()

    return None


def qgis_server_version():
    """ Try to guess the QGIS Server version. """
    # Using ldd :(
    exec_path = os.getenv('QGIS_EXEC_PATH')
    if not exec_path:
        exec_path = "/usr/lib/cgi-bin/qgis_mapserv.fcgi"

    output = subprocess.run(["ls", exec_path], capture_output=True)

    if output.returncode != 0:
        print(f"{exec_path} is not found, not possible to determine QGIS version. Try QGIS_EXEC_PATH")
        return None

    output = subprocess.run(
        ["ldd", exec_path],
        capture_output=True,
        text=True,
    )
    output = [i for i in output.stdout.split('\n') if 'qgis' in i]
    return parse_ldd(output)



"""
Metadata about the package to easily retrieve information about it.
See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

from datetime import date

__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__title_clean__",
    "__uri__",
    "__version__",
    "__version_info__",
]

__author__ = "Étienne Trimaille"
__copyright__ = "2021 - {0}, {1}".format(date.today().year, __author__)
__email__ = "etrimaille@3liz.com"
__license__ = "GNU General Public License v3.0"
__summary__ = "Tool for downloading/managing QGIS plugins from CLI."
__title__ = "QGIS Plugin Manager"
__title_clean__ = "".join(e for e in __title__ if e.isalnum())
__uri__ = "https://github.com/Gustry/qgis-plugin-manager"

# This string might be updated on CI on runtime with a proper semantic version name with X.Y.Z
__version__ = "0.1.0"

if "." not in __version__:
    # If __version__ is still not a proper semantic versioning with X.Y.Z
    # let's hardcode 0.0.0
    __version__ = "0.0.0"

__version_info__ = tuple(
    [
        int(num) if num.isdigit() else num
        for num in __version__.replace("-", ".", 1).split(".")
    ]
)



from typing import NamedTuple


class Plugin(NamedTuple):
    name: str = None
    description: str = None
    version: str = None
    qgis_minimum_version: str = None
    homepage: str = None
    prerelease: str = None
    file_name: str = None
    icon: str = None
    author_name: str = None
    download_url: str = None
    uploaded_by: str = None
    create_date: str = None
    update_date: str = None
    experimental: str = None
    deprecated: str = None
    tracker: str = None
    repository: str = None
    tags: str = None
